@@ -9,6 +9,7 @@ use crate::ImageInfo;
 use crate::ImageView;
 use crate::WindowId;
 use crate::WindowProxy;
+use glam::Vec3;
 use glam::{Affine2, Vec2};
 
 /// Internal shorthand for window event handlers.
@@ -31,9 +32,6 @@ pub struct Window {
 	/// The wgpu surface to render to.
 	pub surface: wgpu::Surface,
 
-	/// The swap chain for the surface.
-	pub swap_chain: wgpu::SwapChain,
-
 	/// The window specific uniforms for the render pipeline.
 	pub uniforms: UniformsBuffer<WindowUniforms>,
 
@@ -81,7 +79,7 @@ impl<'a> WindowHandle<'a> {
 	/// The current window may not be moved or removed through the returned reference.
 	/// In practise, this means that you may not create or destroy any windows.
 	unsafe fn context_mut(&mut self) -> &mut Context {
-		&mut self.context_handle.context
+		self.context_handle.context
 	}
 
 	/// Get a reference to the window.
@@ -133,7 +131,9 @@ impl<'a> WindowHandle<'a> {
 		context_handle
 	}
 
-	/// Get the image info and the area of the window where the image is drawn.
+	/// Get the image info.
+	///
+	/// Returns [`None`] if no image is set for the window.
 	pub fn image_info(&self) -> Option<&ImageInfo> {
 		Some(self.window().image.as_ref()?.info())
 	}
@@ -254,15 +254,35 @@ impl<'a> WindowHandle<'a> {
 	/// The image transformation is applied to the image and all overlays in virtual window space.
 	///
 	/// Virtual window space goes from `(0, 0)` in the top left corner of the window to `(1, 1)` in the bottom right corner.
+	///
+	/// This transformation does not include scaling introduced by the [`Self::preserve_aspect_ratio()`] property.
+	/// Use [`Self::effective_transform()`] if you need that.
 	pub fn transform(&self) -> Affine2 {
 		self.window().user_transform
 	}
 
+	/// Get the full effective transformation from image space to virtual window space.
+	///
+	/// This transformation maps the image coordinates to virtual window coordinates.
+	/// Unlike [`Self::transform()`], this function returns a transformation that include the scaling introduced by the [`Self::preserve_aspect_ratio()`] property.
+	/// This is useful to transform between window coordinates and image coordinates.
+	///
+	/// If no image is set on the window yet, this returns the same transformation as [`Self::transform()`].
+	///
+	/// Virtual window space goes from `(0, 0)` in the top left corner of the window to `(1, 1)` in the bottom right corner.
+	///
+	/// Note that physical pixel locations must be transformed to virtual window coordinates first.
+	pub fn effective_transform(&self) -> Affine2 {
+		self.window().calculate_uniforms().transform
+	}
+
 	/// Set the image transformation to a value.
 	///
 	/// The image transformation is applied to the image and all overlays in virtual window space.
 	///
 	/// Virtual window space goes from `(0, 0)` in the top left corner of the window to `(1, 1)` in the bottom right corner.
+	///
+	/// This transformation should not include any scaling related to the [`Self::preserve_aspect_ratio()`] property.
 	pub fn set_transform(&mut self, transform: Affine2) {
 		self.window_mut().user_transform = transform;
 		self.window_mut().uniforms.mark_dirty(true);
@@ -443,17 +463,20 @@ impl Window {
 	/// Recalculate the uniforms for the render pipeline from the window state.
 	pub fn calculate_uniforms(&self) -> WindowUniforms {
 		if let Some(image) = &self.image {
-			let image_size = image.info().size.as_f32();
+			let image_size = image.info().size.as_vec2();
 			if !self.preserve_aspect_ratio {
 				WindowUniforms::stretch(image_size)
 					.pre_apply_transform(self.user_transform)
 			} else {
-				let window_size = glam::UVec2::new(self.window.inner_size().width, self.window.inner_size().height).as_f32();
+				let window_size = glam::UVec2::new(self.window.inner_size().width, self.window.inner_size().height).as_vec2();
 				WindowUniforms::fit(window_size, image_size)
 					.pre_apply_transform(self.user_transform)
 			}
 		} else {
-			WindowUniforms::no_image()
+			WindowUniforms {
+				transform: self.user_transform,
+				image_size: Vec2::new(0.0, 0.0),
+			}
 		}
 	}
 }
@@ -488,11 +511,11 @@ impl WindowUniforms {
 
 		let w;
 		let h;
-		if ratios[0] >= ratios[1] {
+		if ratios.x >= ratios.y {
 			w = 1.0;
-			h = ratios[1] / ratios[0];
+			h = ratios.y / ratios.x;
 		} else {
-			w = ratios[0] / ratios[1];
+			w = ratios.x / ratios.y;
 			h = 1.0;
 		}
 
@@ -519,15 +542,16 @@ struct Vec2A8 {
 
 #[repr(C, align(16))]
 #[derive(Debug, Copy, Clone)]
-struct Vec2A16 {
+struct Vec3A16 {
 	pub x: f32,
 	pub y: f32,
+	pub z: f32,
 }
 
 #[repr(C)]
 #[derive(Debug, Copy, Clone)]
-struct Mat3x2 {
-	pub cols: [Vec2A16; 3]
+struct Mat3x3 {
+	pub cols: [Vec3A16; 3]
 }
 
 impl Vec2A8 {
@@ -536,14 +560,14 @@ impl Vec2A8 {
 	}
 }
 
-impl Vec2A16 {
-	pub const fn new(x: f32, y: f32) -> Self {
-		Self { x, y }
+impl Vec3A16 {
+	pub const fn new(x: f32, y: f32, z: f32) -> Self {
+		Self { x, y, z }
 	}
 }
 
-impl Mat3x2 {
-	pub const fn new(col0: Vec2A16, col1: Vec2A16, col2: Vec2A16) -> Self {
+impl Mat3x3 {
+	pub const fn new(col0: Vec3A16, col1: Vec3A16, col2: Vec3A16) -> Self {
 		Self {
 			cols: [col0, col1, col2],
 		}
@@ -556,18 +580,21 @@ impl From<Vec2> for Vec2A8 {
 	}
 }
 
-impl From<Vec2> for Vec2A16 {
-	fn from(other: Vec2) -> Self {
-		Self::new(other.x, other.y)
+impl From<Vec3> for Vec3A16 {
+	fn from(other: Vec3) -> Self {
+		Self::new(other.x, other.y, other.z)
 	}
 }
 
-impl From<Affine2> for Mat3x2 {
+impl From<Affine2> for Mat3x3 {
 	fn from(other: Affine2) -> Self {
+		let x_axis = other.matrix2.x_axis;
+		let y_axis = other.matrix2.y_axis;
+		let z_axis = other.translation;
 		Self::new(
-			other.matrix2.x_axis.into(),
-			other.matrix2.y_axis.into(),
-			other.translation.into(),
+			Vec3A16::new(x_axis.x, x_axis.y, 0.0),
+			Vec3A16::new(y_axis.x, y_axis.y, 0.0),
+			Vec3A16::new(z_axis.x, z_axis.y, 1.0),
 		)
 	}
 }
@@ -576,8 +603,8 @@ impl From<Affine2> for Mat3x2 {
 #[repr(C)]
 #[derive(Debug, Copy, Clone)]
 pub struct WindowUniformsStd140 {
-	transform: Mat3x2,
 	image_size: Vec2A8,
+	transform: Mat3x3,
 }
 
 unsafe impl crate::backend::util::ToStd140 for WindowUniforms {
@@ -585,8 +612,8 @@ unsafe impl crate::backend::util::ToStd140 for WindowUniforms {
 
 	fn to_std140(&self) -> Self::Output {
 		Self::Output {
-			transform: self.transform.into(),
 			image_size: self.image_size.into(),
+			transform: self.transform.into(),
 		}
 	}
 }
@@ -602,14 +629,14 @@ pub(super) fn default_controls_handler(mut window: WindowHandle, event: &mut cra
 			let scale = 1.1f32.powf(delta);
 
 			let origin = event.position
-				.map(|pos| pos / window.inner_size().as_f32())
+				.map(|pos| pos / window.inner_size().as_vec2())
 				.unwrap_or_else(|| glam::Vec2::new(0.5, 0.5));
 			let transform = glam::Affine2::from_scale_angle_translation(glam::Vec2::splat(scale), 0.0, origin - scale * origin);
 			window.pre_apply_transform(transform);
 		},
 		WindowEvent::MouseMove(event) => {
 			if event.buttons.is_pressed(crate::event::MouseButton::Left) {
-				let translation = (event.position - event.prev_position) / window.inner_size().as_f32();
+				let translation = (event.position - event.prev_position) / window.inner_size().as_vec2();
 				window.pre_apply_transform(Affine2::from_translation(translation));
 			}
 		},
